/******************************************************************************/
/*                                                                            */
/* Progarm name    : DxPool.c                                                 */
/* Output name     : DxPool.dll                                               */
/* Description     : This Plug-in node will create a shared memory and can    */
/*                   and will be able to read and write data in memory        */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Prafull Kumar                    Date: July 2001                   */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <windows.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */
#include <BipCci.h>		/* Function prototypes for common functions           */
#include "DxPool.h"     /* Specific header for this dll                       */
#include "DxPoolUtil.h" 
#include "node_utils.h" 
#include "pluginMsg.h"  /* Header for Message to event viewer                 */

/*Handle of Shared Memory*/
HANDLE hDxMem=NULL;
/*Handle of Shared Memory View.. Only one view will be loaded per process*/
HANDLE hDxMemView=NULL;
/*Handle of Semaphore*/
HANDLE hSemaphore=NULL;
/*Memroy Pages each of 4096 bytes. default is 1*/
int noOfPoolPages=1;
/*Count of number of users using the view*/
int poolUsers=0;

/******************************************************************************/
/* insertAttributeTableEntry                                                  */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input) - pointer to the current context                    */
/*  nodeName      (input) - node name                                         */
/*  attributeName (input) - attribute name to be inserted into table          */
/*  type          (input) - type of the attribute                             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to an attribute table entry.                                      */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for attribute and link it to the current context.     */
/******************************************************************************/
ATTRIBUTE_TABLE_ENTRY *insertAttributeTableEntry(
												 CONTEXT_STRUCTURE *context
                                                ,wchar_t *nodeName
                                                ,wchar_t *attributeName
												,int type
                                                )
{
  ATTRIBUTE_TABLE_ENTRY *entry;
  int attributeSize;
  /* Allocate storage for the attribute entry */
  entry = malloc(sizeof(ATTRIBUTE_TABLE_ENTRY));

  /* if malloc is successful */
  if (entry)
  {
	/* set this entry to be the new end of list */
    entry->next = 0;
	/* store away the type for this entry */
	entry->type = type;

    /* set the attribute size depending on the type of attribute */
	switch (entry->type)
	{
	  case CCI_VALUE_TYPE_INTEGER:
		attributeSize = 4;
		break;
	  case CCI_VALUE_TYPE_BOOL:
		attributeSize = 4;
		break;
	  case CCI_VALUE_TYPE_CHAR:
        attributeSize = 1;
		/*setting this size to one. as We dont know the size of script in advance
		will allocate the required space when setAttribute will be called.*/
	  default:
		break;
	}

	/* acquire and clear storage for attribute value */
	entry->valuePtr = malloc(attributeSize);
	if (entry->valuePtr)
	  memset(entry->valuePtr, 0, attributeSize);

	/* store away the attribute name */
	wcscpy (entry->name, attributeName);

	/* if a previous element exists in the list, link it to this new one */
	if (context->attributeListPrevious)
	  context->attributeListPrevious->next = entry;
	else if (context->attributeListHead == 0)
	  context->attributeListHead = entry;
	  
	/* save the pointer to the previous element */
	context->attributeListPrevious = entry;
  }

  return(entry);
}

/******************************************************************************/
/* insertInputTerminalListEntry                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to an input terminal list entry.                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for input terminal entry and link it to the current   */
/* context.                                                                   */
/******************************************************************************/

TERMINAL_LIST_ENTRY *insertInputTerminalListEntry(
                                                  CONTEXT_STRUCTURE *context
											     ,wchar_t *terminalName 
                                                 )
{
  TERMINAL_LIST_ENTRY *entry;
  int rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry)
  {
    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
    wcscpy(entry->name, terminalName);

    /* Create terminal and save its handle */
    entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, terminalName);

    /* Link an existing previous element to this one */
    if (context->inputTerminalListPrevious)
	  context->inputTerminalListPrevious->next = entry;
    else if ((context->inputTerminalListHead) == 0)
	  context->inputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->inputTerminalListPrevious = entry;
  }

  return(entry);
}

/******************************************************************************/
/* insertOutputTerminalListEntry                                              */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/*  Pointer to an output terminal list entry.                                 */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for output terminal entry and link it to the current  */
/* context.                                                                   */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insertOutputTerminalListEntry(
                                                   CONTEXT_STRUCTURE *context
												  ,wchar_t *terminalName 
                                                  )
{
  TERMINAL_LIST_ENTRY *entry;
  int rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry)
  {
    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
    wcscpy(entry->name, terminalName);

    /* Create terminal and save its handle */
    entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, terminalName);

    /* Link an existing previous element to this one */
    if (context->outputTerminalListPrevious)
	  context->outputTerminalListPrevious->next = entry;
    else if ((context->outputTerminalListHead) == 0)
	  context->outputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->outputTerminalListPrevious = entry;
  }

  return(entry);
}

/******************************************************************************/
/* getInputTerminalHandle                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named input terminal.                           */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the input terminal list looking for the named input terminal.    */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(
                             CONTEXT_STRUCTURE *context
							,wchar_t *terminalName
                            )
{
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->inputTerminalListHead;
  log("--> getInputTerminalHandle\n");

  /* having set the pointer to the start of the terminal list, traverse down */
  /* and compare each element in the list until the named terminal is found  */
  /* then return the handle to that terminal object, otherwise return null   */
  while (p != 0)
  {
    if (wcscmp(p->name, terminalName) == 0)
	  return(p->handle);
    p = p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/* getOutputTerminalHandle                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named output terminal.                          */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the output terminal list looking for the named output terminal.  */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(
                              CONTEXT_STRUCTURE *context
			                 ,wchar_t *terminalName
                             )
{
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->outputTerminalListHead;
  log("--> getOutputTerminalHandle\n");

  /* having set the pointer to the start of the terminal list, traverse down */
  /* and compare each element in the list until the named terminal is found  */
  /* then return the handle to that terminal object, otherwise return null   */
  while (p != 0)
  {
    if (wcscmp(p->name, terminalName) == 0)
	  return(p->handle);
    p = p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/******************************************************************************/
void *createNodeContext(
						void *factoryObject
					   ,wchar_t *nodeName
					   ,void *nodeObject
					   )
{
  CONTEXT_STRUCTURE *p;
  int rc;
  int isMemoryAlreadyExist;
  /* Allocate a pointer to the local context */
  log("--> createNodeContext\n");
  p		= malloc(sizeof(CONTEXT_STRUCTURE));
  /* If malloc was successful */
  if (p)
  {
    memset (p, 0, sizeof(CONTEXT_STRUCTURE)); /* Clear the context area */
	p->nodeObject	  = nodeObject;
	p->pDxPoolExpressions = 0;

	/* Verify the name of the node is supported by the factory */
	if (wcscmp(nodeName, L"DxPoolNode") == 0)
	{
      /* Build attribute list for the node */
	  insertAttributeTableEntry(p, nodeName, L"DxPoolExpression", CCI_VALUE_TYPE_CHAR);
	  insertAttributeTableEntry(p, nodeName, L"messageStoreEltNotFound", CCI_VALUE_TYPE_BOOL);
	  insertAttributeTableEntry(p, nodeName, L"messageInsertKeyNotFound", CCI_VALUE_TYPE_BOOL);
	  insertAttributeTableEntry(p, nodeName, L"messgaeRestore", CCI_VALUE_TYPE_BOOL);
	  insertInputTerminalListEntry(p, L"in");
	  insertOutputTerminalListEntry(p, L"out");
	  insertOutputTerminalListEntry(p, L"failure");
	  isMemoryAlreadyExist=initHandleOfDxPool();
	  initHandleOfDxPoolView(isMemoryAlreadyExist);
	  if(hDxMem == NULL || hDxMemView == NULL)
	  {
		  /* Error handling if shared Memory handle is NULL */
		  cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "createNodeContext", L"MQSI-DxPool", DXP_CANT_INIT_MEMORY, "MQSeriesIntegrator2",0);
	      cciThrowException(&rc, CCI_CONFIGURATION_EXCEPTION, __FILE__, __LINE__, "createNodeContext", L"MQSI-DxPool", DXP_CANT_INIT_MEMORY, "MQSeriesIntegrator2", (char *)0);
	  }
	}
	else
	{
	  /* Error handling if name of node is not supported by factory */
	}
  }
  else
  {
    /* Error handling if malloc failed */
  }

  /* Return the address of the context to the calling function */
  ++poolUsers;
  log("    poolUsers %d\n",poolUsers);
  return(p);
}

void deleteNodeContext(CciContext* context)
{
	DX_POOL_EXPRESSIONS* p,* tp;
	int iMessageResotre;
    log("--> deleteNodeContext\n");
	--poolUsers;
	if(poolUsers==0)
	{
		iMessageResotre=getBoolAttribute(context,L"messgaeRestore");
		if(iMessageResotre==1)
			storeDxPoolToFile();
		closeHandleOfDxPoolView();
		closeHandleOfDxPool();
		CloseHandle(hSemaphore);
		hSemaphore=NULL;
	}
	p=((CONTEXT_STRUCTURE*)context)->pDxPoolExpressions;
	while(p) {
	  free(p->valuePtr);
	  tp=p->next;
	  free(p);
	  p=tp;
	}
	log("    finish.\n");
}

/******************************************************************************/
/* getAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attributeName (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/*  buffer        (output) - address of buffer into which attribute value is  */
/*                           copied                                           */
/*  bufferSize    (input)  - length of buffer                                 */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Gets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a report request is received that causes a retrieval    */
/* of the value of a node attribute. The broker will have verified that the   */
/* attribute name is valid for the node.                                      */
/******************************************************************************/
int getAttribute(
                 void *context
				,wchar_t *attrName
				,wchar_t *buffer
				,int bufsize      
                )
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;
  while (p != 0)
  {
    if (wcscmp(p->name, attrName) == 0)
	{
      switch(p->type)
	  {
        case CCI_VALUE_TYPE_INTEGER:
          _itow(*((int *)p->valuePtr), buffer, 10); 
          break;
        case CCI_VALUE_TYPE_BOOL:
		  if (*((int *)(p->valuePtr)))
		  {
			  wcscpy(buffer, L"true");
		  }
		  else
		  {
			  wcscpy(buffer, L"false");
		  }
          break;
		case CCI_VALUE_TYPE_CHAR:
	      memcpy(buffer, p->valuePtr, min(MAX_STRING_SIZE, bufsize));
        default:
          break;
      }

      /* Normal return */
      return(0);
    }
    p = p->next;
  }

  /* Unknown attribute */
  return(1); 
}

/******************************************************************************/
/* getBoolAttribute                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attrName      (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - bool attribute not set                                                */
/*  1 - bool attribute is set                                                 */
/* -1 - bool attribute not found                                              */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Gets the value of an boolean attribute                                     */
/******************************************************************************/
int getBoolAttribute(void *context,wchar_t *attrName)
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;
  int val=-1;
  while (p != 0)
  {
    if (wcscmp(p->name, attrName) == 0)
	{
      switch(p->type)
	  {
        case CCI_VALUE_TYPE_BOOL:
		  val=*((int *)(p->valuePtr));
          break;
        default:
          break;
      }

      /* Normal return */
      return(val);
    }
    p = p->next;
  }

  /* Unknown attribute */
  return(-1); 
}

/******************************************************************************/
/* getAttributePointer                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attrName      (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/* Description:                                                               */
/*                                                                            */
/* Gets the Pointer of the Data of an Attirbute                               */
/******************************************************************************/
void* getAttributePointer(
                 void *context
				,wchar_t *attrName
                )
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;
  log("--> getAttributePointer\n");
  while (p != 0)
  {
    if (wcscmp(p->name, attrName) == 0)
	{
      /* Normal return */
      return(p->valuePtr);
    }
    p = p->next;
  }

  /* Unknown attribute */
  return(NULL); 
}

/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/******************************************************************************/
int setAttribute(
                 void *context
				,wchar_t *attributeName
				,wchar_t *attributeValue
                )
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;
  int exLen,rc;
  void *data;
  //log("--> setAttribute\n");
  //log("    Name [%S] Value [%S]\n",attributeName,attributeValue);

  while (p != 0)
  {
    if (wcscmp(p->name, attributeName) == 0)
	{
      switch(p->type)
	  {
        case CCI_VALUE_TYPE_INTEGER:
          *((int *)p->valuePtr) = _wtoi(attributeValue);
          break;
        case CCI_VALUE_TYPE_BOOL:
		  if (!wcscmp(attributeValue, L"false"))
		  {
			*((int *)(p->valuePtr)) = 0;
		  }
		  else
		  {
			*((int *)(p->valuePtr)) = 1;
			if(wcscmp(attributeName,L"messgaeRestore")==0)
						initDxPoolFromFile();
		  }
          break;

        case CCI_VALUE_TYPE_CHAR:
		  if(wcscmp(attributeName,L"DxPoolExpression")==0)
		  {
			  exLen=wcslen(attributeValue);
			  data = malloc(sizeof(wchar_t)*exLen);
			  //free(p->valuePtr);
			  p->valuePtr=data;
			  memcpy(p->valuePtr, attributeValue, 2*exLen);
			  rc=processDxPoolscript(context);
			  if(rc!=0)// return rc;
			  {
				  log("    DxPoolExpression is wrong..\n");
			      cciThrowException(&rc, CCI_CONFIGURATION_EXCEPTION, __FILE__, __LINE__, "setAttribute", L"MQSI-DxPool", 10, "MQSeriesIntegrator2", (char *)0);
			  }
		  }
		  else
	          memcpy(p->valuePtr, attributeValue, MAX_STRING_SIZE);
          break;
        default:
          break;
      }

      /* Normal return */
      return(0);
    }
    p = p->next;
  }

  /* Error...attribute not found */
  return(1);
}

/******************************************************************************/
/* getAttributeName                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context    (input)  - pointer to the current context                      */
/*  index      (input)  - index of the attribute for which the name is to be  */
/*                        retrieved                                           */
/*  buffer     (output) - address of buffer into which attribute name is      */
/*                        copied                                              */
/*  bufferSize (input)  - length of buffer                                    */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Returns the name of an attribute by a given index. It is invoked by the    */
/* broker when it requires the names of attributes supported by a particular  */
/* instance of a node. The function must gurantee to return the attribute     */
/* in a known, defined order and to return the attribute name represented by  */
/* the index parameter.                                                       */
/*                                                                            */
/******************************************************************************/
int getAttributeName(
                     void* context
					,int index
					,wchar_t* buffer
					,int bufferSize
                    )
{
  int count = 0;
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;

  /* having set the pointer to the start of the attribute list, traverse down*/
  /* and compare each element in the list until the indexed element is found */
  /* then return the name of the attribute in the supplied buffer.           */
  while (p != 0)
  {
    if (count == index)
	{
      wcscpy(buffer, p->name);
      return(0);
    }
    count++;
    p = p->next;
  }

  /* Invalid index */
  return(1);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/******************************************************************************/
void evaluate(
              void* context
			 ,void* destinationList
			 ,void* exceptionList
			 ,void* message
             )
{
  CCI_EXCEPTION_ST exceptionStructure;
  void* sendTerminalObject;
  int rc;
  int retValue = 1;
  int attributeBufSize = 0;
  wchar_t attributeBuffer;
  wchar_t terminal[10];
  wchar_t* pAttributeBuffer = &attributeBuffer;
  char functionName[] = "evalute()";
  CciMessage*      outMsg;
  CciMessage*      sendMsg;
  CciElement*      inRootElement;
  CciElement*      outRootElement;
  char *logbuffer=NULL;
  log("--> evaluate\n");

  inRootElement = cniRootElement(&rc, message);
  /* Create a new message for output */
  outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));

  /* Get the root element of the output message */
  outRootElement = cniRootElement(&rc, outMsg);

  /* Copy the contents of the input message to the output message */
  cniCopyElementTree(&rc, inRootElement, outRootElement);

  retValue=runDxPoolscript(context,message,destinationList,exceptionList,outMsg,&logbuffer);
  log("    retValue of runDxPoolScript %d\n",retValue);

  /* propagate the message to the output terminal */
  /* get the handle to the output terminal */
  if(retValue == DXP_SUCCESS)  
  {
	  wcscpy (terminal,L"out");
	  sendMsg=outMsg;
  }
  else
  {
	  wcscpy (terminal,L"failure");
	  sendMsg=message;
	  if(logbuffer==NULL)
	  {
		cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSI-DxPool", retValue, "MQSeriesIntegrator2",0);
	  }
	  else
	  {
		  cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSI-DxPool", retValue, "MQSeriesIntegrator2",logbuffer,0);
		  free(logbuffer);
		  logbuffer=NULL;
	  }
  }

  sendTerminalObject = getOutputTerminalHandle(context, terminal);

  /* check to see if terminal exists and is attached */
  if (sendTerminalObject)
  {
	if (cniIsTerminalAttached(&rc, sendTerminalObject))
	{
	  if (rc == CCI_SUCCESS)
	  {
		if(retValue == DXP_SUCCESS)  
			cniFinalize(&rc, outMsg, 0);
        retValue = cniPropagate(
			                    &rc
							   ,sendTerminalObject
							   ,destinationList
							   ,exceptionList
							   ,sendMsg
							   );
		log("    cniPropogate.. retValue=%d rc=%d\n",retValue,rc);
		if (rc != CCI_SUCCESS)
		{
		  if (rc == CCI_EXCEPTION)
		  {
			cciGetLastExceptionData(&rc, &exceptionStructure);
			cciRethrowLastException(0);
		  }
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", DXP_MESSAGE_NOT_PROPAGATED, "Message not propagated", (char *)0);
		}
		else
		{
		  //cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", BIP2623, "Message propagated", (char *)0);
		}
	  }
	  else
	  {
        if (rc == CCI_EXCEPTION)
		{
          cciRethrowLastException(0);
		}
		else
		{
		  /* some other error, better log it */
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSI-DxPool", DXP_MESSAGE_NOT_PROPAGATED, "Message not propagated", (char *)0);
        }
	  }
	}
  }
  else
  {
	/* Terminal does not exist */
    cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSI-DxPool", DXP_TERMINAL_NOT_ATTACHED, "terminal not attached",terminal,0);
  }

  cniDeleteMessage(0, outMsg);
  return;
}

/******************************************************************************/
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Address of factory object.                                                */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine                                        */
/*                                                                            */
/******************************************************************************/
void __declspec(dllexport) *bipGetMessageflowNodeFactory()
{
  static char* functionName="bipGetMessageflowNodeFactory()";
  void* factoryObject;
  wchar_t factoryName[] = L"DxPoolFactory";
  CCI_EXCEPTION_ST exception_st;
  int rc;
  static CNI_VFT vftable;

  /* Setup virtual function table with addresses of implementation functions */
  memset(&vftable, 0, sizeof(vftable));
  vftable.iFpCreateNodeContext     = createNodeContext;
  vftable.iFpDeleteNodeContext     = deleteNodeContext;
  vftable.iFpGetAttributeName      = getAttributeName;
  vftable.iFpSetAttribute          = setAttribute;
  vftable.iFpGetAttribute          = getAttribute;
  vftable.iFpEvaluate              = evaluate;
		
  /* Output a log message */
  cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2",
         0, "Plugin <DxPoolFactory> loading", (char *)0);

  /* Create the Node Factory for this plug-in */
  factoryObject = cniCreateNodeFactory(0, factoryName);
  if (factoryObject == CCI_NULL_ADDR)
  {
    if (rc == CCI_EXCEPTION)
	{
      cciGetLastExceptionData(&rc, &exception_st);
      /* Any local error handling may go here */
      cciRethrowLastException(&rc);
    }
    /* Any further local error handling can go here */
  }
  else
  {
    /* Define a node type supported by our factory */
    cniDefineNodeClass(&rc, factoryObject, L"DxPoolNode", &vftable); 
    if (rc != CCI_SUCCESS)
	{
      /* Error Handling */
    }

  }

  /* Return address of this factory object to the broker */
  return(factoryObject);
}

/******************************************************************************/
/* initHandleOfDxPool                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is to create the shared Memory segment. If the Shared Memory */
/* already exist, it gets the handle of that segment.                         */
/*                                                                            */
/******************************************************************************/
int initHandleOfDxPool()
{
	char poolSize[20];
	int rc;
	log("--> initHandleOfDxPool hDxMem = 0x%X\n",hDxMem);
	
	hSemaphore=CreateSemaphore(NULL,1,1,DX_POOL_SEMAPHORE_NAME);
	if(hSemaphore==NULL)
	{
		log("    hSemaphore is Null\n");
		CloseHandle(hSemaphore);
	}
	if(hDxMem!=NULL)
		return 1;

	if (getenv("DXPOOL_MEMORY_PAGES") != NULL) 
	{
		strcpy(poolSize, getenv("DXPOOL_MEMORY_PAGES")); 
		noOfPoolPages=atoi(poolSize);
		if(noOfPoolPages<=0 || noOfPoolPages>5000)
			noOfPoolPages=1;
		log("    DXPOOL_MEMORY_PAGES %s %d\n",poolSize,noOfPoolPages);
	}

	hDxMem = OpenFileMapping(FILE_MAP_ALL_ACCESS, 0,DX_POOL_MEMORY_NAME);
	if(hDxMem!=NULL)
	{
	    log("    OpenFileMapping hDxMem = 0x%X\n",hDxMem);
		return 1;
	}

	sprintf(poolSize,"%d",noOfPoolPages);
    cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "initHandleOfDxPool", L"MQSI-DxPool", DXP_MEMORY_SIZE, "MQSeriesIntegrator2",poolSize,0);
	hDxMem = CreateFileMapping((HANDLE)0xFFFFFFFF,NULL,PAGE_READWRITE,0,DX_POOL_MEMORY_SIZE*noOfPoolPages,DX_POOL_MEMORY_NAME);
    log("    hDxMem = 0x%X\n",hDxMem);
	if(hDxMem == NULL)
	{
		CloseHandle(hDxMem);
	}
	return 0;
}

/******************************************************************************/
/* initHandleOfDxPoolView                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  MemoryExist       (input) - is Memory alreay exist.                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is to get to get the handle of the view of Shared Memory     */
/*                                                                            */
/******************************************************************************/
void initHandleOfDxPoolView(int MemoryExist)
{
    log("--> initHandleOfDxPoolView hDxMemView = 0x%X\n",hDxMemView);
	if(hDxMemView!=0)		/*if hDxMemView is already initialized*/
		return;
	hDxMemView = MapViewOfFile(hDxMem, FILE_MAP_ALL_ACCESS, 0, 0, 0); 
	log("    hDxMemView = 0x%X\n",hDxMemView);
	if (hDxMemView == NULL)  
	{ 
		CloseHandle(hDxMemView);
	} 
	if(!MemoryExist)
		initDxPoolMemory();
}

/******************************************************************************/
/* initDxPoolMemory                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* to initialize the shared Memory segment                                    */
/*                                                                            */
/******************************************************************************/
void initDxPoolMemory()
{
	DX_POOL_CONTROL p;
    log("--> initDxPoolMemory\n");
	p.poolSize			= DX_POOL_MEMORY_SIZE*noOfPoolPages;
	p.DxPoolDataBlocks	= 0;
	p.DxPoolUsedSize	= sizeof(DX_POOL_CONTROL);
    WaitForSingleObject(hSemaphore,INFINITE);
	memcpy(hDxMemView,&p,sizeof(DX_POOL_CONTROL));
	ReleaseSemaphore(hSemaphore,1,NULL);
}

/******************************************************************************/
/* log                                                                        */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to log the data in file c:\DxPool.log                                 */
/*                                                                            */
/******************************************************************************/
void log(char *strFormat, ...)
{
	static FILE *fp = NULL;

	va_list	list;
	va_start(list, strFormat);

	if (fp == NULL)
	{
		fp = fopen("C:\\DxPool.log", "a");
		if (fp == NULL)
		{
			return;
		}
	}

	vfprintf(fp, strFormat, list);
	fflush(fp);
	va_end(list);
}

/******************************************************************************/
/* closeHandleOfDxPool                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to close the handle of Shared Memory handle                           */
/*                                                                            */
/******************************************************************************/
void closeHandleOfDxPool()
{
	log("--> closeHandleOfDxPool hDxMem %X\n",hDxMem);
	if(hDxMem!=NULL)
		CloseHandle(hDxMem); 
	hDxMem=NULL;
	log("    closeHandleOfDxPool .. now onward hDxMem is NULL\n");
}

/******************************************************************************/
/* closeHandleOfDxPoolView                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to close the handle of Shared Memory view                             */
/*                                                                            */
/******************************************************************************/
void closeHandleOfDxPoolView()
{
	log("--> closeHandleOfDxPoolView hDxMemView %X\n",hDxMemView);
	if(hDxMemView!=NULL)
		UnmapViewOfFile(hDxMemView);
	hDxMemView=NULL;
	log("    closeHandleOfDxPoolView .. now onward hDxMemView is NULL\n");
}

/******************************************************************************/
/* processDxPoolscript                                                       */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/* !0 - failure, wrong script                                                 */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to to store each line of the script in structure DxPoolExpressions    */
/* and check that that the script should be correctness else throw exception  */
/*                                                                            */
/******************************************************************************/
int processDxPoolscript(CONTEXT_STRUCTURE* context)
{
	int charindex=0,oldcharindex=0,times=0,temp=0;
	int rc;
	char data[60];
	CciChar* poolScript = getAttributePointer(context,L"DxPoolExpression");
	char*  scriptline;
	log("--> processDxPoolscript\n");//-[%S] length-%d\n",poolScript,CciCharLen(poolScript));
	/*****Varification of Script***********/
	if( (rc=CciCharIndexOf(poolScript,L'*',1))>=0 ||
		(rc=CciCharIndexOf(poolScript,L'/',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'\\',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'+',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'-',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'"',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'\'',1)>=0) )
	{
		sprintf(data,"wrong keyword(Not Supported) '%c' used in Script",poolScript[rc]);
		cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "setAttribute", L"MQSI-DxPool", DXP_SCRIPT_ERROR, "MQSeriesIntegrator2",data, (char *)0);
		return 100;
	}
	/*****Varification of Script***********/

	for(;;){
		oldcharindex=charindex;
		charindex=CciCharIndexOf(poolScript,chSemiComa,++times);
		if(charindex==-1) break;
		//log("    charindex/old %d %d\n",charindex,oldcharindex);
		scriptline=malloc(charindex-oldcharindex+1);
		memset(scriptline,0,charindex-oldcharindex+1);
		if(oldcharindex!=0) temp=1;
		/* to store data in array in char format */
		UTOASCIICharNCpy(scriptline,poolScript+oldcharindex+temp,charindex-oldcharindex-temp);
		rc=insertDxScript(context,scriptline);
		if(rc!=0) return rc;
		//log("    scriptline [%s] length %d\n",scriptline,strlen(scriptline));
	}
	return 0;
}

int insertDxScript(CONTEXT_STRUCTURE* context,char *pScript)
{
  DX_POOL_EXPRESSIONS* entry;
  DX_POOL_EXPRESSIONS* p;
  int rc;
  unsigned int i; char* c;char* set;
  //check for any errors. remove '\n' '\t'...
  log("--> insertDxScript Script[%s]\n",pScript);
  c=pScript;
  for(i=0;i<strlen(pScript);i++)
  {
	  if(*c=='\n' || *c=='\t' || *c==' ')
	  {
		  set=c;
		  for(;*set=='\n' || *set=='\t' || *set==' ' || *set==0;set++);
		  strcpy(c,set);
	  }
	  c++;
  }
  if(strncmp(pScript,"SET",3)!=0) return 1;
  strcpy(pScript,pScript+3);
  if(strncmp(pScript,"Input",strlen("Input"))==0)
  {
	log("    Input.. %s\n",pScript);
    cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "insertDxScript", L"MQSI-DxPool", DXP_TRY_TO_CHANGE_INPUT_MSG, "MQSeriesIntegrator2",pScript, (char *)0);
	return 2;
  }

  entry = malloc(sizeof(DX_POOL_EXPRESSIONS));

  if (entry)
  {
    entry->next = 0;
	entry->valuePtr=pScript;
	p=context->pDxPoolExpressions;
	if(p==0)
	{
		context->pDxPoolExpressions=entry;
		return 0;
	}
	while(p->next!=0) p=p->next;
	p->next=entry;
	return DXP_SUCCESS;
  }
  return DXP_FAILED;
}

int getOperandsOfPoolScript(char* s,char* l,char* r)
{
	int end;
    log("--> getOperandsOfPoolScript\n");
	end=CharIndexOf(s,'=',1);
	strncpy(l,s,end); 	*(l+end)=0;
	strncpy(r,s+end+1,strlen(s)-end-1);	*(r+strlen(s)-end-1)=0;
	//log("    l=[%s],r=[%s]\n",l,r);
	return 0;
}

/******************************************************************************/
/* runDxPoolscript                                                            */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  message         (input) - address of the input message object             */
/*  exceptionList   (input) - address of the exceptionList message object     */
/*  destinationList (input) - address of the destinationList message object   */
/*  outMsg          (in/out)- address of the output message object            */
/*  logbuff         (output)- address of the to log by CciLog                 */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/* !0 - failure, reason code of failure                                       */
/*                                                                            */
/* Description:                                                               */
/* This script is called to execute all the DxPool Expression.                */
/*                                                                            */
/******************************************************************************/
int  runDxPoolscript(CONTEXT_STRUCTURE* context,CciMessage* message,
					 CciMessage* exceptionList,CciMessage* destinationList,CciMessage* outMsg,char** logbuff)
{
  DX_POOL_DATA   pool;
  DX_POOL_EXPRESSIONS* p;
  CciElement *element;
  char lOperand[MAX_STRING_SIZE];
  char rOperand[MAX_STRING_SIZE];
  int rc,len,alen;
  //CciChar elementValue[256];
  CciChar *elementValue;
  CciChar *ev;
  char* charElementValue=NULL;
  CciMessage* realDests = destinationList;
  CciMessage* realExcs = exceptionList;
  CciMessage* realMsg = message;
  CciElement* foundElt;
  char* pDataToUse=NULL;
  int iNullShmAllow,iNullEltAllow;
  log("--> runDxPoolscript\n");
  iNullShmAllow=getBoolAttribute(context,L"messageStoreEltNotFound");
  iNullEltAllow=getBoolAttribute(context,L"messageInsertKeyNotFound");

  pool.data=NULL;

  p=context->pDxPoolExpressions;
  while(p) {
	  log("    COMMAND ->[%s]\n",p->valuePtr);
	  getOperandsOfPoolScript(p->valuePtr,lOperand,rOperand);
	  if(strncmp(rOperand,"DxPool",6)==0)
	  {
		rc=getDxPoolDataBlock(&pool,rOperand);
		if(iNullEltAllow==0 && rc!=0)
		{
			*logbuff=strdup(rOperand);
			return DXP_KEY_NOT_FOUND;
		}
		if(iNullEltAllow==1 && rc!=0)
		{
			pool.data=malloc(1);
			strcpy(pool.data,"");
		}
		pDataToUse=pool.data;
	  }
	  else
	  {
		log("    calling NodeGetValue\n");
		if(strncmp(rOperand,"Output",strlen("Output"))==0)
		{
			strcpy(rOperand,rOperand+strlen("Output"));
			//log("Output.. %s\n",rOperand);
			element=NodeGetValue(rOperand,outMsg,realExcs,realDests);
		}
		else 
		{
			if (strncmp(rOperand,"Input",strlen("Input"))==0)
				strcpy(rOperand,rOperand+strlen("Input"));
			//log("Input.. %s\n",rOperand);
			element=NodeGetValue(rOperand,realMsg,realExcs,realDests);
		}
		if(element)
		{
			len=256;
			elementValue=malloc(sizeof(CciChar)*(len+1));
			alen=cniElementCharacterValue(&rc,element,(CciChar*)elementValue,len);
			if(alen>len)
			{
				free(elementValue);
				elementValue=malloc(sizeof(CciChar)*(alen+1));
				cniElementCharacterValue(&rc,element,(CciChar*)elementValue,alen);
			}
			log("    cniElementCharacterValue [%S]\n",elementValue);
			charElementValue=UnicodeToASCII(elementValue);
			pDataToUse=charElementValue;
			free(elementValue);
		}
		else
		{
			log("    runDxPoolscript-element not found\n");
			if(iNullShmAllow)
			{
				charElementValue=malloc(1);
				strcpy(charElementValue,"");
				pDataToUse=charElementValue;
			}
			else
			{
				*logbuff=strdup(rOperand);
				return DXP_ELEMENT_NOT_FOUND;
			}
		}
	  }
	  if(strncmp(lOperand,"DxPool",6)==0)
	  {
		//log("    calling storeDxPoolDataBlock\n");
		pool.dataKey=lOperand;
		pool.data=pDataToUse;
		rc=storeDxPoolDataBlock(&pool);

		if(rc!=0) return rc;
		//should return 1, if no store .. memory is full..??
	  }
	  else
	  {
		if(strncmp(lOperand,"Output",strlen("Output"))==0)
		{
			strcpy(lOperand,lOperand+strlen("Output"));
			//log("lOperand Output.. %s\n",lOperand);
		}
		log("    calling NodeAddOrGetElement\n");
		foundElt=NodeAddOrGetElement(lOperand,outMsg,exceptionList,destinationList);
		if(foundElt!=NULL)
		{
			//log("    element is not NULL... calling cniSetElementCharacterValue\n");
			ev=ASCIIToUnicode(pDataToUse);
			log("    data [%S]\n",ev);
		    cniSetElementCharacterValue(&rc,foundElt,ev,wcslen(ev));
			//log("    going to free the ev !!\n");
			if(ev!=NULL) {free(ev);ev=NULL;}
		}
		else
		{
			log("    element is NULL\n");
			free(pDataToUse);
			*logbuff=strdup(lOperand);
			return DXP_CANT_CREATE_ELEMENT;
		}
	  }//main else
	  free(pDataToUse);
	  p=p->next;
  }
  return 0;
}

/******************************************************************************/
/* getDxPoolDataBlock                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  d        (input)  - pointer to DX_POOL_DATA, to retrieve the information  */
/*  inkey    (input)  - key value to get                                      */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown key                                                  */
/*                                                                            */
/* Description:                                                               */
/*  Search the Shared Memory to fill the structure and return it back         */
/******************************************************************************/
int getDxPoolDataBlock(DX_POOL_DATA* d,char* inkey)
{
	int i;
	char tKey[MAX_KEY_SIZE];
	void *p;
	char* data=NULL;
	DX_POOL_CONTROL c;
	log("--> getDxPoolDataBlock\n");
    WaitForSingleObject(hSemaphore,INFINITE);

	memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
	p=(char*)hDxMemView+sizeof(DX_POOL_CONTROL);
	//log("    getDxPoolDataBlock-%d blocks\n",c.DxPoolDataBlocks);
	for(i=0;i<c.DxPoolDataBlocks;i++)
	{
		memcpy(d,p,3*sizeof(int));
		//log("    getDxPoolDataBlock keylength=%d datalength=%d\n",d->keyLength,d->dataLength);
		//log("    getDxPoolDataBlock 0x%X %d\n",(char*)p+3*sizeof(int),min(d->keyLength,MAX_KEY_SIZE));
		memcpy(tKey,(char*)p+(3*sizeof(int)),min(d->keyLength,MAX_KEY_SIZE));
		//log("    tKey [%s]--nextIncrement %d\n",tKey,d->totalLength);
		if(strcmp(tKey,inkey)==0)
		{
			//log("    getDxPoolDataBlock found the Key in Memory.\n");
			data=malloc(d->dataLength);
			memcpy(data,(char*)p+3*sizeof(int)+d->keyLength,d->dataLength);
			d->data=data;
			//log("    getDxPoolDataBlock [%s]\n",d->data);
			ReleaseSemaphore(hSemaphore,1,NULL);
			return DXP_SUCCESS;
		}
		p=(char*)p+d->totalLength;
	}
	log("    Key Not FOUND?\n");
	ReleaseSemaphore(hSemaphore,1,NULL);
	return DXP_KEY_NOT_FOUND;
}

/******************************************************************************/
/* storeDxPoolDataBlock                                                       */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  d        (input)  - pointer to DX_POOL_DATA, to store the information     */
/*                                                                            */
/* Returns:                                                                   */
/*   NULL                                                                     */
/*                                                                            */
/* Description:                                                               */
/*  Stores the data to the shared Memory.                                     */
/******************************************************************************/
int storeDxPoolDataBlock(DX_POOL_DATA* d)
{
	DX_POOL_CONTROL c;
	DX_POOL_DATA data;
	int i,size=0;
	char tKey[MAX_KEY_SIZE];
	void *p;
	
	log("--> storeDxPoolDataBlock\n");
	size=sizeof(DX_POOL_CONTROL);

    WaitForSingleObject(hSemaphore,INFINITE);
	
	memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
	//log("    ####control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
	p=(char*)hDxMemView+sizeof(DX_POOL_CONTROL);
	//check if the key already exist.
	//log("    DxPoolDataBlocks=%d\n",c.DxPoolDataBlocks);
	for(i=0;i<c.DxPoolDataBlocks;i++)
	{
		memcpy(&data,p,3*sizeof(int));
		//log("    storeDxPoolDataBlock TOTAL=%d keylength=%d datalength=%d\n",data.totalLength,data.keyLength,data.dataLength);
		//log("    storeDxPoolDataBlock 0x%X %d\n",(char*)p+3*sizeof(int),min(data.keyLength,MAX_KEY_SIZE));
		memcpy(tKey,(char*)p+(3*sizeof(int)),min(data.keyLength,MAX_KEY_SIZE));
		size+=data.totalLength;
		if(strcmp(tKey,d->dataKey)==0)
		{
			log("    Key already exist.. ?\n");
			memcpy((char*)p,(char*)p+data.totalLength,c.DxPoolUsedSize-size);
			//log("    #### 0x%X 0x%X %d\n",p,(char*)p+data.totalLength,(DX_POOL_MEMORY_SIZE*noOfPoolPages)-size);
			c.DxPoolDataBlocks--;
			c.DxPoolUsedSize-=data.totalLength;
			//log("    #### %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
			break;
		}
		p=(char*)p+data.totalLength;
	}
	d->keyLength=strlen(d->dataKey)+1;
	d->dataLength=strlen(d->data)+1;
	d->totalLength=3*sizeof(int)+d->keyLength+d->dataLength;
	log("    Used Pool Memory %d out of %d\n",c.DxPoolUsedSize+d->totalLength,(DX_POOL_MEMORY_SIZE*noOfPoolPages));
	if(c.DxPoolUsedSize+d->totalLength>(DX_POOL_MEMORY_SIZE*noOfPoolPages))
	{
		ReleaseSemaphore(hSemaphore,1,NULL);
		return DXP_MEMORY_FULL;
	}
	memcpy((char*)hDxMemView+c.DxPoolUsedSize,d,sizeof(int)*3);
	memcpy((char*)hDxMemView+c.DxPoolUsedSize+sizeof(int)*3,d->dataKey,d->keyLength);
	memcpy((char*)hDxMemView+c.DxPoolUsedSize+sizeof(int)*3+d->keyLength,d->data,d->dataLength);
	c.DxPoolUsedSize+=d->totalLength;
	c.DxPoolDataBlocks++;
	//log("    %d %d %d %s %s\n",d->totalLength,d->keyLength,d->dataLength,d->dataKey,d->data);
	memcpy(hDxMemView,&c,sizeof(DX_POOL_CONTROL));
	ReleaseSemaphore(hSemaphore,1,NULL);
	return 0;
}

/******************************************************************************/
/* storeDxPoolToFile                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None                                                                      */
/*                                                                            */
/* Returns:                                                                   */
/*  0 - success                                                               */
/*  1 - failure                                                               */
/*                                                                            */
/* Description:                                                               */
/*  Stores the data of shared Memory to a file c:\DxPoolData.bin              */
/******************************************************************************/
int storeDxPoolToFile()
{
  DX_POOL_CONTROL c;
  FILE *fp=NULL;
  int ret;
  log("--> storeDxPoolToFile\n");
  if(hDxMemView==NULL) return DXP_FAILED;
	
  fp = fopen("c:\\DxPoolData.bin","wb");
  if(fp==NULL) return DXP_FAILED;

  WaitForSingleObject(hSemaphore,INFINITE);
  memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
  //log("    ####control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
  ret=fwrite(hDxMemView,1,c.DxPoolUsedSize,fp);
  if(ret!=c.DxPoolUsedSize)
  {
	  //take action.. if not able to read full file
  }
  log("    return of fwrite %d\n",ret);
  fclose(fp);
  ReleaseSemaphore(hSemaphore,1,NULL);
  cciLog(&ret, CCI_LOG_ERROR, __FILE__, __LINE__, "storeDxPoolToFile", L"MQSI-DxPool", DXP_MEMORY_SAVED_TO_FILE, "MQSeriesIntegrator2", (char *)0);
  return DXP_SUCCESS;
}

/******************************************************************************/
/* initDxPoolFromFile                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None                                                                      */
/*                                                                            */
/* Returns:                                                                   */
/*  0 - success                                                               */
/*  1 - failure                                                               */
/*                                                                            */
/* Description:                                                               */
/*  fills the shared Memory from file c:\DxPoolData.bin                       */
/******************************************************************************/
int initDxPoolFromFile()
{
  FILE *fp=NULL;
  int lSize;
  int ret;
  DX_POOL_CONTROL c;
  int rc;
  log("--> initDxPoolFromFile\n");

  fp = fopen("c:\\DxPoolData.bin", "rb");
  if(fp==NULL) return DXP_FAILED;
  fseek (fp , 0 , SEEK_END);
  lSize = ftell (fp);
  rewind (fp);
  if(lSize <= sizeof(DX_POOL_CONTROL)) //if file size is less then control size
  {
	fclose(fp);
    return DXP_FAILED;
  }
  ret=fread(&c,1,sizeof(DX_POOL_CONTROL),fp);
  rewind (fp);
  if(ret!=sizeof(DX_POOL_CONTROL)) // if not able to read the req bytes
  {
	fclose(fp);
	return DXP_FAILED;
  }
  if(c.DxPoolUsedSize!=lSize) //if file size if not equal to the used size
  {
	fclose(fp);
	return DXP_FAILED;
  }

  WaitForSingleObject(hSemaphore,INFINITE);
  memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
  //log("    #control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
  if(c.DxPoolUsedSize==sizeof(DX_POOL_CONTROL)) //if the pool is not used??
  {
	if(c.DxPoolDataBlocks<DX_POOL_MEMORY_SIZE*noOfPoolPages) // only if the file size is less then the pool
	{
		ret=fread(hDxMemView,1,lSize,fp);
		log("    return of fwrite %d\n",ret);
		if(ret==lSize)
	      cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "initDxPoolFromFile", L"MQSI-DxPool", DXP_MEMORY_INIT_FROM_FILE, "MQSeriesIntegrator2", (char *)0);
	}
  }
  fclose(fp);
  ReleaseSemaphore(hSemaphore,1,NULL);

  return DXP_SUCCESS;
}
